/***************************************************************************************************
 Project : Loupeznik - krotitel 230 loupezniku v 10-ti zasuvkach 
 Date    : 7.6.2007
 Author  : Jan Matuska, honzamat@atlas.cz                          
 Company : www.starhill.org
 Funkce  : ovladaci logika odpojovani 10-ti zasuvek pres relatka pomoci RS232 ze serveru
 ---------------------------------------------------------------------------------------------------
 Chip type                : ATmega8L
 Clock frequency          : 4.0 MHz (internal RC)
 Memory model             : Small
 Internal SRAM size       : 1024 B
 EEPROM memory            : 256 B   
 Compiler                 : avr-gcc 4.1.2 (WinAVR 20070525rc2) 
 Fuses (log.0=programmed) : BOOTSZ1, BOOTSZ2, SUT0, CKSEL3, CKSEL2
***************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "loupeznik.h"
#include "messages.h"


// globalni promenne v EEPROM
uint8_t ee_relcon[RELAYS_COUNT]   EEPROM_SECTION ;   // mapovaci pole na relatka

// globalni promenne pro UART, volatile=je videt v preruseni
volatile unsigned char rs232enter = 0;              // globalni flag prijateho radku na RS232 (\n)            
volatile unsigned char rx_buffer[RX_BUFFER_SIZE];   // buffer prijatych znaku na RS232 

// glogalni citac pro generovani zpozdeni pomoci preruseni a delicka na blikani LEDky
volatile unsigned int counter = 0; 
volatile unsigned char ledscaler = 0x08; 


// =================================================================================================        
// R U Z N E
// -------------------------------------------------------------------------------------------------
// preruseni kazdou 1ms na generovani zpozdeni  4e6/4e3=1e3  65536-4e3=61536=0xF060 do TCNT1
ISR(TIMER1_OVF_vect) {
  TCNT1H=0x0F0;
  TCNT1L=0x60;
  if (counter) {
    counter--;
    CLRWDT;
  }
}
// -------------------------------------------------------------------------------------------------
// cekaci smycka po kroku 1ms
void delay(unsigned int cas) {
  cli();        // prirazeni neni atomicka operace, zakaz preruseni 
  counter=cas;
  sei();
  while(counter) asm("nop");
}
// -------------------------------------------------------------------------------------------------
// test ASCII znaku zda je argument cislo od 0-9
unsigned char is_numeric(unsigned char arg)  {
  if ((arg>='0') && (arg<='9')) return 1; 
  return 0;
}
// -------------------------------------------------------------------------------------------------
// umocnovani zakladu 10 a exponentem exp, tedy 10^exp
unsigned int decpower(unsigned char exp) {
  unsigned int out=1;
  while(exp--) out *= 0x0A;
  return out;
}
// -------------------------------------------------------------------------------------------------   
// vezme pocet znaku v rx_bufferu od pocize start a prevede na cisleny char
// value=vystupni hodnota, error_code=1 pro OK 0 pro failed (overflow, no-numeric apod.)
unsigned char rxb2char(unsigned char *value, unsigned char start, unsigned char count) {
  unsigned char i=0;
  unsigned int tmp=0;
  while(count--) tmp += decpower(count) * (rx_buffer[start-(i++)]-48);
  if(tmp<=255) {
    *value = (unsigned char) tmp;
    return 1;
  } else return 0;
}
// ------------------------------------------------------------------------------------------------- 
// obsluha OVF preruseni Timeru1 pro blikani stavove LEDky. Uzasna cinnost, ze? :-))
ISR(TIMER0_OVF_vect) {
  if(!(ledscaler>>=1)) {
    if(!(PORTB&LED)) PORTB|=LED; else PORTB&=~LED;
    ledscaler=0x08; 
  }
}
// ------------------------------------------------------------------------------------------------- 
// projede pole relcon v EEPROM a pokud je v nem hodnota vetsi nez je pocet relatek, tak upravi 
// hodnotu na nejakou akceptovatelnou.  
void test_relcon(void) {
  unsigned char i=RELAYS_COUNT;
  while(i--) 
    if (eeprom_read_byte(&ee_relcon[i])>=RELAYS_COUNT) eeprom_write_byte(&ee_relcon[i],i+1); 
}  
             


// =================================================================================================        
// R S - 2 3 2
// -------------------------------------------------------------------------------------------------  

// preruseni po prichodu noveho znaku na USART     
// novy byte zapise do pozice 0, ostatni posune o 1 vyse = FIFO fornta
// po prichodu \n nastavi priznak enteru - je zpracovan v main()
ISR(USART_RXC_vect) {
  unsigned char status,data,i;
  status=UCSRA;
  data=UDR;
  if ((status & (FRAMING_ERROR | PARITY_ERROR | DATA_OVERRUN))==0) {
     if((data=='\n')) { rs232enter=1; }
     if(data>=32 && data<=126) {
       i=RX_BUFFER_SIZE-1;
       while(i>0) {
         rx_buffer[i]=rx_buffer[i-1];
         i--;
      }
     rx_buffer[0]=data;
     }
  }
}
// -------------------------------------------------------------------------------------------------         
// odeslani znaku na USART
// ceka na vyprazneni TX bufferu a zapise do nej novy znak  
void putchar1(char c) {
  while (!(UCSRA & DATA_REGISTER_EMPTY));
  UDR=c;
}       
// -------------------------------------------------------------------------------------------------    
// odesle na terminal znaky konce radku a navratu na zacatek
void lfcr(void) {
  putchar1('\n');    //send lf - 10
  putchar1('\r');    //send cr - 13
}
// -------------------------------------------------------------------------------------------------  
// vypsani znakoveho retezce na terminal z flash pameti  
void write(char *sss){
  char k;     
  while ((k=pgm_read_byte(sss++))>0) {
    putchar1(k);  
  } 
  lfcr();
} 
// -------------------------------------------------------------------------------------------------  
// vypsani znakoveho retezce na terminal  
void writestr(char *sss){
  char k;       
  while ((k=(*sss++))>0) {
     putchar1(k);   
  }    
} 
// -------------------------------------------------------------------------------------------------  
// vypsani unsigned charu na terminal  
void print_char(unsigned char arg, unsigned char rad){
  unsigned char i=3;
  while(i--) {
    if (i<=(rad-1))  putchar1(48 + (arg / decpower(i)));
    arg %= decpower(i);
  }     
}
// -------------------------------------------------------------------------------------------------  
// otestuje, zdali se cmd nachazi na count pozici bufferu
// a zda-li je za cmd pouze numericky argument
unsigned char cmdtest(unsigned char cmd, unsigned char count){
  if(cmd!=rx_buffer[count]) return 0;
  while(count--) {
    if(!is_numeric(rx_buffer[count])) {
      write(UART_NUMERROR);
      return 0;
    }
  }  
  return 1;
}
// ------------------------------------------------------------------------------------------------- 
// vypis mapovaci tabulky relatek
void rs232_showrelcon(void) {  
  unsigned char i=1;
  writestr("rele | port");
  lfcr();
  writestr("-----------");
  lfcr();
  do {
    print_char(i,2);
    writestr(" ----> ");
    print_char(eeprom_read_byte(&ee_relcon[i-1]),2);
    lfcr();
  } while(++i<=RELAYS_COUNT);
  lfcr();
}
// ------------------------------------------------------------------------------------------------- 
// zapis mapovaciho zaznamu
void rs232_allocrelcon(void) {
  unsigned char pp,rr;
  rxb2char(&pp,1,2);
  rxb2char(&rr,3,2);
  if((rr>RELAYS_COUNT)||(pp>RELAYS_COUNT)||(!pp)||(!rr)) { 
    write(UART_PROVF);
    return;
  }
  print_char(rr,2);
  writestr(" ----> ");
  print_char(pp,2);
  lfcr();
  eeprom_write_byte(&ee_relcon[rr-1],pp);
  write(UART_PRSAVED);
}
// ------------------------------------------------------------------------------------------------- 
// provede reset
void rs232_reset(void) {
  unsigned char port,cas,i=RELAYS_COUNT;
  rxb2char(&port,3,2);
  rxb2char(&cas,1,2);
  if((port>RELAYS_COUNT)||(!port)) { 
    write(UART_PROVF);
    return;
  }  
  write(UART_RESETBEG);
  while(i--) if(eeprom_read_byte(&ee_relcon[i])==port) relay_on(i+1);
  LED_FAST;
  while(cas--) {
    delay(1000); 
    putchar1(0x8);            // srandicka s countdownem na terminalu
    putchar1(0x8);
    print_char(cas,2);      
  }
  putchar1(0x8);
  putchar1(0x8);
  LED_SLOW;
  relay_off();
  write(UART_RESETOK); 
}  
// -------------------------------------------------------------------------------------------------
// rozdeleni akci jednotlivym prikazum
void rs232_command(void) {   
  unsigned char i=RX_BUFFER_SIZE;
  if(cmdtest('P',0)) write(UART_ALIVE); 
  if(cmdtest('H',0)) write(HELP); 
  if(cmdtest('L',0)) rs232_showrelcon();
  if(cmdtest('A',4)) rs232_allocrelcon();
  if(cmdtest('R',4)) rs232_reset();
  while (i--) rx_buffer[i]=0;
  rs232enter=0;     
}



// =================================================================================================        
// R E L E  
// -------------------------------------------------------------------------------------------------  
// sepne rele dle parametru rele. Vicenasobnym volanim lze sepnout vice rele
// pozor na proudovy odber rele, aby to zdroj rozchodil.
void relay_on(unsigned char rele) {  
  if(rele==1)  RELE_01_ON;
  if(rele==2)  RELE_02_ON;
  if(rele==3)  RELE_03_ON;
  if(rele==4)  RELE_04_ON;
  if(rele==5)  RELE_05_ON;
  if(rele==6)  RELE_06_ON;
  if(rele==7)  RELE_07_ON;
  if(rele==8)  RELE_08_ON;
  if(rele==9)  RELE_09_ON;
  if(rele==10) RELE_10_ON;
}
// ------------------------------------------------------------------------------------------------- 
// odpoji vsechny relatka (pusti do nich stavu)
void relay_off(void) {  
  RELE_01_OFF;
  RELE_02_OFF;
  RELE_03_OFF;
  RELE_04_OFF;
  RELE_05_OFF;
  RELE_06_OFF;
  RELE_07_OFF;
  RELE_08_OFF;
  RELE_09_OFF;
  RELE_10_OFF;
}                              
// ================================================================================================= 





// The Story Begins here...
int main(void) {
  // port B
  PORTB=0x00;
  DDRB=0x03;    // 0000 0011
  // port C
  PORTC=0x00;
  DDRC=0x0FC;   // 1111 11000
  // port D
  PORTD=0x0;
  DDRD=0x0FF;   // 1111 1111
  // Timer 0 - blikani LEDky, uzasna cinnost :-)))
  TCCR0=0x05;   // nastaveni delicky pro blikani LED resi makro o radek nize
  LED_SLOW;
  // Timer/Counter 1 
  TCCR1A=0x00;
  TCCR1B=0x01;  // bez preddelicky   
  TCNT1H=0x0F0;
  TCNT1L=0x60;
  OCR1AH=0x00;
  OCR1AL=0x00;
  OCR1BH=0x00;
  OCR1BL=0x00;
  // Povoleni preruseni Timer1 overflow a Timer0 overflow
  TIMSK=0x05;  // preruseni Overflow timer1 a Overflow timer0
  // UART  RX: on, TX: on, 9600 baud, 8bit, 1 stop bit, no parity
  UCSRA=0x00;
  UCSRB=0x98; 
  UCSRC=0x86;
  UBRRL=25;  // speed 9600
  UBRRH=0x00;
  // Analog Comparator initialization - nepouzivan
  // ACSR=0x80;   // snizeni spotreby vypnutim komparatoru
  // SFIOR=0x00;
  // Global enable interrupts
  sei();
  //  Zapnuti HW hlidaciho cokla  
  CLRWDT;
  WDTCR=0x18;
  WDTCR=0x0F;
  // Global enable interrupts
  sei();
  test_relcon();
  write(WELCOME);
  while(1) { 
    CLRWDT;
    if(rs232enter==1) rs232_command();
  }
}

